// This source file is free for any kind of use commercial or non by any party.

// Author: Conor Stokes
// Purpose: AABB tree operations

#include "shareddefs.h"
#include "vector.h"
#include "geom.h"
#include "aabbtree.h"

// constructs this aabb tree node from a triangle list and creates its children recursively
// note node box is initialised to the first triangle's box
CAABBNode::CAABBNode(list<CTriangle>& trilist,UINT32 depth,const Heuristic& heurdata) : m_box(trilist) {
	// test our build heuristic - if passes, make children
	if (depth<heurdata.maxdepth&& trilist.size()>heurdata.mintricnt && 
		(trilist.size()>heurdata.tartricnt || m_box.ErrorMetric()>heurdata.minerror)) {

		list<CTriangle>::const_iterator triitr; // iterator for looping through trilist 
		list<CTriangle> tribuckets[2]; // buckets of triangles
		UINT32 longaxis=m_box.LongestAxis(); // longest axis
		vec3d geoavg; // geometric average - midpoint of ALL the triangles

		// go through all tris and calculate the average of the midpoints
		for (triitr=trilist.begin(); triitr!=trilist.end(); triitr++)
			geoavg=geoavg + triitr->MidPoint()*(1.0f/trilist.size());			 

		// bucket tris based on their midpoint's side of the geo average in the longest axis
		for (triitr=trilist.begin(); triitr!=trilist.end(); triitr++)
			tribuckets[geoavg[longaxis]>triitr->MidPoint()[longaxis]].push_back(*triitr);

		// create new children using the buckets
		m_children[0]=new CAABBNode(tribuckets[0],depth+1,heurdata);
		m_children[1]=new CAABBNode(tribuckets[1],depth+1,heurdata);	
	
	} else { // otherwise the build heuristic failed, this is 
		
		// set the first child to null (identifies a leaf)
		m_children[0]=0;
		
		// copy triangle list
		m_triangles=trilist;
	}	
}

// destructor
CAABBNode::~CAABBNode() { 
	// if a note a leaf node, delete the children.
	if (m_children[0]) { 
		delete m_children[0]; 
		delete m_children[1]; 
	}
}

// renders tris that collide with the sphere
void CAABBNode::RenderCollideTris(const CSphere& testsphere) const {
	// if no intersection with the node box, don't continue
	if (testsphere.Intersects(m_box)) {
		
		m_box.Render();

		// is not a leaf
		if (m_children[0]) {
			// recurse to children
			m_children[0]->RenderCollideTris(testsphere);
			m_children[1]->RenderCollideTris(testsphere);

		} else { // is a leaf

			list<CTriangle>::const_iterator triitr;

			// loop through and render triangles, if they intersect.
			for (triitr=m_triangles.begin(); triitr!=m_triangles.end(); triitr++)
				if (testsphere.Intersects(*triitr)) 
					triitr->Render();
		}
	}
}
