// This source file is free for any kind of use commercial or non by any party.

// Author: Conor Stokes
// Purpose: Simple geometry primitive operations

#include "shareddefs.h"
#include "vector.h"
#include "geom.h"

// construct an AABB from a list of triangles.
CAABB::CAABB(const list<CTriangle>& triangles) {
	// do nothing if no triangles in the list
	if (triangles.empty())
		return;

	list<CTriangle>::const_iterator triitr;	

	// copy aabb of first triangle 
	Cpy(triangles.front().TriAABB());	
	
	// encompass all other triangles in the list
	for (triitr=triangles.begin(); triitr!=triangles.end(); triitr++) 
		Encompass(triitr->TriAABB());
}

// distance squared to a point from the box (arvos algorithm)
float CAABB::DistSqrd(const vec3d& point) const { 
	float dst=0; 

	//for each component, find the point's relative position and the distance contribution
	for (UINT32 ii=0; ii<3; ii++) 
		dst+=(point[ii] < m_bounds[0][ii]) ? SQR(point[ii] - m_bounds[0][ii]) : 
			 (point[ii] > m_bounds[1][ii]) ? SQR(point[ii] - m_bounds[1][ii]) : 0.0f;

	return dst; 
}

// longest axii of the bounding box
UINT32 CAABB::LongestAxis() const {
	vec3d boxdim(m_bounds[1]-m_bounds[0]);
	
	UINT32 la=0; // longest axis
	float lav=0.0f; // longest axis length
	
	// for each dimension  
	for (UINT32 ii=0; ii<3; ii++) {
		// check if its longer
		if (boxdim[ii]>lav) {
			// store it if it is
			la=ii;
			lav=boxdim[ii];
		}
	}
	
	return la;
}

// make this box encompass the current box as well as this one
void CAABB::Encompass(const CAABB& encbox) {
	m_bounds[0]=m_bounds[0].Min(encbox.m_bounds[0]);
	m_bounds[1]=m_bounds[1].Max(encbox.m_bounds[1]);
} 

// get the aabb for this triangle (used for making group aabb)
CAABB CTriangle::TriAABB() const {
	// new aabb with min as minimum of 3 verts and maximum max of 3 verts
	return CAABB(m_verts[0].Min(m_verts[1]).Min(m_verts[2]),
				 m_verts[0].Max(m_verts[1]).Max(m_verts[2]));
} 

// distance squared to a point from the tri
// note SMOD is just a specialised case of a modulation 
float CTriangle::DistSqrd(const vec3d& point) const {
	UINT32 ii;
	UINT32 edgesidecount=0; // edge side total
	UINT32 edgeside[3]; // edge side (inside/outside)
	vec3d hpnorm; // half plane normal
	vec3d edgevec[3]; // edge vectors
	
	// for all edges
	for (ii=0; ii<3; ii++) {
		// create an edge vector
		edgevec[ii]=m_verts[ii]-m_verts[SMOD((ii+1),3)];
		
		// make a normal for the plane through the edge perpindicular to the triangle
		hpnorm=m_trinorm.Cross(edgevec[ii]);
		
		// test the side on which our point lies relative to this plane
		edgeside[ii]=hpnorm.Dot(point)<=(hpnorm.Dot(m_verts[ii]));
		edgesidecount+=edgeside[ii];
	}
	
	// this is closest to a point
    if (edgesidecount==1) { 
		for (ii=0; ii<3; ii++) { 
			
			// find the edge this is inside and the distance to the corresponding vert
			if (edgeside[ii])  
				return (point-m_verts[SMOD((ii+2),3)]).MagSqr();
		}
	}

	// this is closest to an edge (line segment distance)
	if (edgesidecount==2) {	
		for (ii=0; ii<3; ii++) {// find the edge this is not inside
			if (!edgeside[ii]) { // then return the distance to the edge
		
				// point is closest the first end of the line segment
				if (edgevec[ii].Dot(point)>edgevec[ii].Dot(m_verts[ii]))
					return (point-m_verts[ii]).MagSqr();
						
				// point is closest the far end of the line segment
				if (edgevec[ii].Dot(point)<edgevec[ii].Dot(m_verts[SMOD(ii+1,3)]))
					return (point-m_verts[SMOD(ii+1,3)]).MagSqr();
						
				// point is clocest the line part of the segment
				return edgevec[ii].Cross(m_verts[ii]-point).MagSqr()/edgevec[ii].MagSqr();
			}
		}
	}
			
	// falls back to distance to the plane, so return it
	return SQR(m_trinorm.Dot(point)-m_tri_d);
} 
